#
#*******************************************************************************
#* Copyright (C) 2018, International Business Machines Corporation. 
#* All Rights Reserved. *
#*******************************************************************************
#

# Import the SPL decorators
from streamsx.spl import spl
from streamsx.ec import get_application_configuration
from streamsx.ec import is_active
from bundleresthandler.wmlbundlecontroller import WmlBundleController
# standard python imports
import re, os
import sys
import logging
import json
import time
import threading
import pickle



#define tracer and logger
#logger for error which should and can! be handled by an administrator
#tracer for all other events that are of interest for developer
tracer = logging.getLogger(__name__)
logger = logging.getLogger("com.ibm.streams.log")


# Defines the SPL namespace for any functions in this module
def spl_namespace():
    return "com.ibm.streams.wml"




######################################################
# callable class just to be given as output handler
# taking the result list of handler and using
# SPL operators submit() function.
# It is storing the reference to the SPL operator
# to reach the correct submit().
######################################################
class output_class():
    def __init__(self, output_object):
        self._output_object = output_object
    def __call__(self, results):
        #with self._output_object._output_lock:
        tracer.debug("Start output_function")
        for index,result_list in enumerate(results):
            tracer.debug("Start result submission ")
            if index == 0:
                for list_element in result_list:
                    self._output_object.submit('result_port',{'__spl_po':memoryview(pickle.dumps(list_element))})
            elif index == 1:
                for list_element in result_list:
                    self._output_object.submit('error_port',{'__spl_po':memoryview(pickle.dumps(list_element))})
            else:
                tracer.error("Internal error: More result lists generated than supported. ")



@spl.primitive_operator(output_ports=['result_port','error_port'])
class WMLOnlineScoring(spl.PrimitiveOperator):
    """Providing the functionality to score incomming data with a model of any of the WML supported frameworks.
    The members __init__ and __call__ of this class will be called when topology application is submitted to the Streams runtime.
    So the thread of the runtime is the one putting the input tuple into the queue.
    
    It is designed to be used in a topology function to consume a stream of incoming tuples and 
    produce a stream of outgoing tuples with scoring results or in case of scoring errors a stream of
    tuples with error indication
    """
    def __init__(self, deployment_guid, 
                       field_mapping,		
                       wml_credentials , 
                       space_guid, 
                       expected_load,         # depricated
                       queue_size, 
                       threads_per_node,
                       single_output,
                       node_count,
                       bundle_size):
        """Instantiates a WMLOnlineScoring object at application runtime (Streams application runtime container).
        
        It creates a WML client connecting to WML service with provided credentials and
        retrieves the URL of the deployment which should be used for scoring.        
        It creates the threads which handle the requests towards the scoring deployment.
        These threads will consume tuples in the input queue, which is filled by the __call__ member.
        """
        tracer.debug("__init__ called")

        #self._output_lock = threading.Lock()
       
        # Configure the specific controller class to be used
        self._controller = WmlBundleController(
                        # controller base class arguments
                        expected_load = expected_load,      # depricated
                        queue_size = queue_size, 
                        threads_per_node = threads_per_node,
                        single_output = single_output,
                        node_count = node_count,
                        field_mapping = field_mapping,
                        output_function = output_class(self),
                        bundle_size = bundle_size,
                        # wml specific controler argumnets
                        deployment_guid = deployment_guid, 
                        wml_credentials = wml_credentials, 
                        space_guid = space_guid
                        ) 


        # type of input stream, set with first received tuple
        self._is_python_object_stream = None

        tracer.debug("__init__ finished")
        return
        

    def __enter__(self):
        tracer.debug("__enter__ called")
        self._controller.prepare()
        tracer.debug("__enter__ finished")

    def all_ports_ready(self):
        tracer.debug("all_ports_ready() called")
        self._controller.run()
        tracer.debug("all_port_ready() finished, sending threads started")
        return self._controller.finish()
    


    @spl.input_port()
    def process(self, **python_tuple):
        """It is called for every tuple of the input stream.
        The tuple will be given to a controller which may
        block and force backpressure on the up-stream.
        """
        # python_tuple is either a k/v element with key '__spl_po' indicating that the content v is a pickled Python dictionary.
        # Or python_tuple is a dictionary containing the k/v representation of a structured schema
        # (nameTuple or SPL schema). 
        # So we need to load it back in an object with pickle.load(<class byte>) from memoryview
        # we receive here as the pickled python object is put in a SPL tuple <blob __spl_po> and
        # SPL type blob is on Python side a memoryview object
        # python tuple is choosen as input type, which has tuple values in sequence of SPL tuple
        # we have control over this SPL tuple and define it to have single attribute being a blob 
        # the blob is filled from topology side with a python dict as we want to work on a dict
        # as most comfortable also when having no defined attribute sequence anymore

        input_tuple=None
        if self._is_python_object_stream is True:
            input_tuple = pickle.loads(python_tuple['__spl_po'].tobytes())
        elif self._is_python_object_stream is False:
            input_tuple = python_tuple
        # only entered at first tuple, next time only the above two are checked
        elif "__spl_po" in python_tuple:
            input_tuple = pickle.loads(python_tuple['__spl_po'].tobytes())
            self._is_python_object_stream = True
        else:
            input_tuple = python_tuple
            self._is_python_object_stream = False
            
        # controller.process can block calling thread here until input_tuple can be stored 
        self._controller.process_data(input_tuple)


    def __exit__(self, exc_type, exc_value, traceback):
        tracer.debug("__exit__ called")
        self._controller.stop()
        tracer.debug("__exit__ finished, triggered all sub threads to stop")
    
    
            
    
    
    
    





#
#*******************************************************************************
#* Copyright (C) 2018, International Business Machines Corporation. 
#* All Rights Reserved. *
#*******************************************************************************

# Bundle
from .wmlbundleresthandler import WmlBundleRestHandler
from .bundlecontroller import BundleController

# WML specific imports
from watson_machine_learning_client import WatsonMachineLearningAPIClient

# standard python imports
import logging
import json
import time
import threading
import pickle
import sys


#define tracer and logger
#logger for error which should and can! be handled by an administrator
#tracer for all other events that are of interest for developer
tracer = logging.getLogger(__name__)


class WmlBundleController(BundleController):
    def __init__(self, deployment_guid = None, 
                       wml_credentials = None, 
                       space_guid = None, 
                       **kwargs
                       ):

        tracer.debug("__init__ called")
        
        ######################################################
        # initialize the controller base class with arguments 
        # and add the handler class to be used
        ######################################################
        kwargs["handler_class"] = WmlBundleRestHandler
        super().__init__(**kwargs)
        ######################################################
        # initialize this controller sub class with its own
        # special arguments
        ######################################################
        self._deployment_guid = deployment_guid
        self._wml_credentials = json.loads(wml_credentials)
        self._deployment_space = space_guid
        ######################################################
        # set specialized handler class class variables 
        # we know at this place which specialized class
        # we use as we set it above
        # the handler base class class variables are set by
        # our own base class
        ######################################################
        #tracer.debug("Handler class: ", self._handler_class)
        self._handler_class.wml_client = self._create_wml_client()
        self._handler_class.deployment_guid = self._deployment_guid 

        tracer.debug("__init__ finished")
        return


    def _change_deployment_node_number(self):
        return


    def _get_deployment_status(self):
        return

    
    def _create_wml_client(self):
        tracer.debug("Creating WML client")
        wml_client = WatsonMachineLearningAPIClient(self._wml_credentials)
        # set space before using any client function
        wml_client.set.default_space(self._deployment_space)
        tracer.debug("WML client created")
        return wml_client




# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2020
   
   
from watson_machine_learning_client import WatsonMachineLearningAPIClient
from watson_machine_learning_client.wml_client_error import ApiRequestFailure

import logging
import numpy   
tracer = logging.getLogger(__name__)   
logger = logging.getLogger("com.ibm.streams.log")   

from .bundleresthandler import BundleRestHandler   
   
   
_STREAMSX_MAPPING_ERROR_ = "Mapping error: "
   
class WmlBundleRestHandler(BundleRestHandler):

    ########################################
    # public controllable class varibales
    ########################################
    wml_client = None
    deployment_guid = None
        
        
    def __init__(self,handler_index):
        super().__init__(handler_index)

        
    def preprocess(self):
        """WML specific implementation,
        One has to know the fields the model requires as well as the schema of input data.
        Those data is defined through the mapping configuration, set as class variable
        by outer application.
        
        Depending on the framework one need to provide the fields of the names or not.
        
        The payload format for WML scoring is a list of dicts containing "fields" and "values".
        "fields" is a list of fieldnames ordered as the model it requires
        "values" is a 2 dimensional list of multiple scoring data sets, where each set is a list of ordered field values 
        [{"fields": ['field1_name', 'field2_name', 'field3_name', 'field4_name'], 
        "values": [[value1, value2, value3, value4],[value1, value2,  value3, value4]]}]

        In case of a single_array_only input there may be no "fields" infomation needed, as the position in array 
        is sufficient.
        
        To indicate from application side that there is no field mapping necessary, the special mapping
        element [{"model_field":"__array__","tuple_field":"<name_of_tuple_field>"}]  is defined.
        This mapping has to be the only one, additional other mapping entries are not valid.
        The type of the tuple_field can be List, numpy.ndarray or pandas.DataFrame. 
        numpy.ndarray and pandas.DataFrame have to be converted to List. (may be changed in future 
        so that the wml client will do conversation.)
        Actually our implementation has to take care for a single line numpy.ndarray as scoring input.
        
        We don't support:
        - multi dimensional ndarray as scoring input for a single tuple
        - pandas.DataFrame as scoring for a single tuple
          DF are used commonly as column labeled two dimensional tabular data, which doesn't match
          single tuple input scenario but batch input scenario which we don't support.
        
        In case of invalid scoring input WML online scoring will reject the whole bundle with "invalid input" 
        reason without indicating which of the many inputs was wrong!!!
        """
        # this is a sample where all fields are required and are anytime in the input tuple
        # model fields have to be in order/sequence as expected by the model

        # keep this assert as long as we don't support optional fields
        assert self.allow_optional_fields is False
        
        # clear payload list
        self._payload_list = []      
        actual_input_combination ={'fields':[],'values':[]}
        
        for index,_tuple in enumerate(self._data_list):
            ###################################
            # field names and values from tuple
            # handling fields names is already preparation
            # for optional field support and resulting
            # different payload content
            ###################################
            tuple_values = []
            tuple_fields = []
            tuple_error = ""
            tuple_is_valid = True
            for field in self.field_mapping:
                if field['tuple_field'] in _tuple and _tuple[field['tuple_field']] is not None:
                    tuple_field_value = _tuple[field['tuple_field']]
                    # Handle the case that a single numpy array / List is the scoring input 
                    # special mapping '__array__'has to be the only mapping
                    # no field name used, ndarray needs 
                    if field['model_field'] == '__array__':
                        if isinstance(tuple_field_value, numpy.ndarray) and len(self.field_mapping) is 1:
                            tuple_values = list(tuple_field_value)
                            break
                        elif isinstance(tuple_field_value, list)  and len(self.field_mapping) is 1:
                            tuple_values = tuple_field_value
                            break
                        else:
                            tuple_is_valid = False
                            break
                    # the normal case, multiple mapping entries and using field name 
                    # tuple field value
                    else:
                        tuple_values.append(tuple_field_value)
                        tuple_fields.append(field['model_field'])

                #########################################################
                # optional fields are actually not supported
                # uncomment next prepared code lines in case we support 
                # it in all code parts
                #########################################################
                #elif self.allow_optional_fields:
                #    if field['is_mandatory']:
                #        tuple_is_valid = False
                #        break
                
                else:
                    tuple_is_valid = False
                    break

            if not tuple_is_valid:            
                tuple_error = "input field: " + field['tuple_field']

            ######################
            # store tuple status
            ######################
            if tuple_is_valid: 
                self._status_list[index]["mapping_success"] = True
            else:
                self._status_list[index]["mapping_success"] = False
                self._status_list[index]["message"] = _STREAMSX_MAPPING_ERROR_ + tuple_error
                continue            

            
            #######################
            # add payload for tuple
            #######################    
            if actual_input_combination['fields'] == tuple_fields:
                #same fields as before, just add further values
                actual_input_combination['values'].append(list(tuple_values))
            elif len(actual_input_combination['values']) is 0:  
                # first tuple
                actual_input_combination['fields']=tuple_fields
                actual_input_combination['values']=[list(tuple_values)]
            else:
                tuple_is_valid = False
                tuple_error = " optional fields are not allowed"
                # following code is preparation for support of multiple field combinations
                # in case of supported optional fields
                #
                #close and store last fields/values combination in final _payload_list
                #except for the first valid tuple being added
                #if len(actual_input_combination['values']) > 0 : self._payload_list.append(actual_input_combination) 
                ##create new field/value combination
                #actual_input_combination['fields']=tuple_fields
                #actual_input_combination['values']=[list(tuple_values)]

                        
        #after last tuple store the open field/value combination finally in payload_list
        if len(actual_input_combination['values']) > 0:
            # remove empty fields element resulting from single array as input 
            if len(actual_input_combination['fields']) is 0:
                actual_input_combination.pop('fields')
            self._payload_list.append(actual_input_combination)
    
    
    def synch_rest_call(self):
        rest_success = True
        error_message = None
        
        try:
            if len(self._payload_list) > 0:
                self._rest_response = self.wml_client.deployments.score(self.deployment_guid,meta_props={'input_data':self._payload_list})
        except ApiRequestFailure as err:
            """REST request returns 
            400 incase something with the value of 'input_data' is not correct
            404 if the deployment GUID doesn't exists as REST endpoint
                    
            score() function throws in this case an wml_client_error.ApiRequestFailure exception
            with two args: description [0] and the response [1]
            use response.status_code, response.json()["errors"][0]["code"], response.json()["errors"][0]["message"]
                   
            The complete payload is rejected in this case, no single element is referenced to be faulty
            As such we need to write the complete payload to invalid_tuples being submitted to 
            error output port
            """
            tracer.error("WML API error description: %s",str(err.args[0]))
            logger.error("WMLOnlineScoring: WML API error: %s",str(err.args[0]))
            #print("WML REST response headers: ",err.args[1].headers)
            #print("WML REST response statuscode: ",err.args[1].status_code)
            #print("WML REST response code: ",err.args[1].json()["errors"][0]["code"])
            #print("WML REST response message: ",err.args[1].json()["errors"][0]["message"])
            #add the complete local tuple list to invalid list
            #TODO one may think about adding an error indicator if tuple is rejected from mapping function
            #or from scoring as part of a scoring bundle
            #because the predictioon for whole bundle failed, the complete local_list is invalid
            rest_success = False
            error_message = str(err.args[0])
        except:
            tracer.error("Unknown exception: %s", str(sys.exc_info()[0]))
            logger.error("WMLOnlineScoring: Unknown exception: %s", str(sys.exc_info()[0]))
            #because the predictioon for whole bundle failed, the complete local_list is invalid
            rest_success = False
            error_message = str(sys.exc_info()[0])
            
        if rest_success:
            for item in self._status_list:
                if item["mapping_success"]:
                    item["score_success"] = True
        else:
            self._rest_response = None
            for item in self._status_list:
                item["score_success"] = False
                if item["message"] is None:
                    item["message"] = "WML API error: " + error_message

        #tracer.debug("WMLOnlineScoring: Worker %d got %d predictions from WML model deployment!", self._handler_index, len(self._rest_response['predictions'][0]['values']))
    
        return len(self._rest_response['predictions']) # number of prediction response bundles
        

    def postprocess(self):

        #scoring REST call had error, no result to process
        #just the error fields have to be provided
        if self._rest_response is None:
            for index, item in enumerate(self._status_list):
                self._result_list[index] = {"PredictionError": item["message"]}
            return 
            
        #take the tuples from local list in sequence, sequence is same as the 
        #sequence of prediction 'values' as input was generated in sequence of the _data_list
        #there is no reference from input to prediction except the position in sequence
        #use output mapping function or just add the raw result to tuple
        #for later separation and processing
        #each prediction contains model result 'fields' and one or more 'values' lists
        #one value list for each scoring set
        # only data with successful mapping was added in payload and gets response data
        for prediction in self._rest_response['predictions']:
            response_index = 0 
            for data_index,item in enumerate(self._status_list):
                if item["mapping_success"] :
                    self._result_list[data_index] = {"Prediction" : dict(zip(prediction['fields'],prediction['values'][response_index]))}
                    response_index += 1 
                else:
                    self._result_list[data_index] = {"PredictionError": item["message"]}
                    




# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2020
   
'''

class ProcessStorage stores 
    - a set of elements of the input data
    - all status and result information related to each of single input element


- it is needed for copy and hold a slice of input data 
- map mining fields (parameter mapping_dict)
    - mapping status may be: successful | invalid_data
    - mapping status has to be stored for each element
- mapping generates scoring request payload for each element
- payload may be devided by 'field' combinations if optional fields are allowed (backlog feature)
- payload dividing depends ordered and unordered submission behavior

'''   

import time
import sys   
import logging



   
tracer = logging.getLogger(__name__)   
   
class BundleRestHandler():

    max_copy_size = 100 
    '''max input items to be copied from input queue, to be set by using application, defaults to 100'''
    source_data_list = None
    '''Reference to the source where data bundles should be read from.'''
    input_list_lock = None
    '''lock for the input queue, this lock is used by different threads acessing the queuue, to be set by using application'''
    field_mapping = None
    '''list with input data attribute to mining model field mapping'''
    keep_data_order = True
    '''For future use: The actual version doesn't support unordered output
    Defines if output data shall be send in same order as input data. 
    This has only impact if optional fields are allowed.
    '''
    single_output = True
    ''' Defines if all results should be written to one output or two (success,error) '''
    allow_optional_fields = False
    ''' For future use: The actual version doesn't support optional fields.
    '''
    bundle_counter = 0
    ''' Counter incremented by each copy of a bundle from source. Reference to an outside variable.
    '''
    next_bundle_to_sent = 0
    ''' Number (derived from bundle_counter) of next bundle to be send. Needed in case keep_data_order = True
    Reference to an outside variable.
    '''
    output_function = None
    ''' Reference of the output function to be used. Depending on single_output setting it has to support
    one or two parameters of type list'''
    
    def __init__(self, handler_index):
        '''ProcessStorage is created build up by copying oldest input data from input queue
        and deleting them from this queue.
        '''
        tracer.debug("Enter init")
        assert self.input_list_lock is not None
        assert self.source_data_list is not None
        assert self.field_mapping is not None
        assert self.output_function is not None
        #if keep_data_order:
        #    assert self.bundle_counter is not None
        #    assert self.next_bundle_to_send is not None
        
        #####################################################################
        # thread related members
        #
        #####################################################################
        self._handler_index = handler_index
        self._run = True
        
        #####################################################################
        # Data processing members
        #
        #####################################################################
        self._data_size = 0
        self._bundle_number = None
        self._data_list = []            # don't change this, this is the original data, filled by copy()
        self._status_list = []          # processing status of several steps, if one step is not successful, next are not done
                                        # will get same size as _data_list
        self._payload_list = []         # REST payload list, each element is one payload, generated by preprocess()
        self._rest_response = []        # holds the REST response, REST errors are reflected in _status_list
        self._result_list = []          # REST result list, needs to have one entry for each data being in payload, result index equals to data index 


    def run(self):
        tracer.debug("Starting handler: %d with threads run method", self._handler_index  )
        
        bundle_tuple_count = 0
        overall_count = 0
        #as long as thread shall not stop
        while self._run:
            bundle_tuple_count = self.copy_from_source()
            tracer.debug("Loop: Thread %d received %d tuples.", self._handler_index, bundle_tuple_count )
            if  bundle_tuple_count > 0:
                self.preprocess()
                self.synch_rest_call()
                self.postprocess()
                self.write_result_to_output()
                overall_count += bundle_tuple_count
                
        tracer.info("Handler %d stopped after %d records", self._handler_index, overall_count)

            
    def stop(self):
        tracer.debug("Stopping handler: %d", self._handler_index  )
        self._run = False


    def copy_from_source(self):        
        self._data_size = 0
        self._bundle_number = None
        self._data_list = []            # don't change this, this is the original data, filled by copy()
        self._status_list = []          # processing status of several steps, if one step is not successful, next are not done
                                        # will get same size as _data_list
        self._payload_list = []         # REST payload list, each element is one payload, generated by preprocess()
        self._rest_response = []        # holds the REST response, REST errors are reflected in _status_list
        self._result_list = []          # REST result list, needs to have one entry for each data being in payload, result index equals to data index 

        # python threading is just sequential processing, staying little longer in lock doesn't matter
        with self.input_list_lock:
            
            # wait blocking with timeout (to allow checking outer run condition) 
            # until list has been filled
            if not self.input_list_lock.wait_for(lambda:len(self.source_data_list) > 0 , 0.1): 
                return 0
        
            #determine size and copy max size or all to local data list
            input_size = len(self.source_data_list)
            #tracer.debug("ProcessStorage (%d) : source_data_list len before copy %d!", self._handler_index, input_size)
            if input_size > 0:
                end_index = int(self.max_copy_size) if input_size >= self.max_copy_size else input_size
                self._data_list = self.source_data_list[:end_index]
                del self.source_data_list[:end_index]
                self._data_size = end_index 
                #tracer.debug("ProcessStorage (%d) :  read %d tuples from input queue with _data_list len %d!", self._handler_index, end_index, len(self._data_list))
                self._bundle_number = self.bundle_counter
                self.bundle_counter += 1
    
                #create the status list at once
                self._status_list = [{"mapping_success":False,"score_success":False,"message":None} for i in range(self._data_size)]
                #create the result list at once
                self._result_list = [None for i in range(self._data_size)]
                #tracer.debug("ProcessStorage (%d) : source_data_list len after copy %d!", self._handler_index, len(self.source_data_list))
                
                #wake up waiting threads inclusive writing threads
                self.input_list_lock.notify()
        return self._data_size                
        
    def get_final_data(self, single_list = True):
        if single_list:
            single_output = [{**data,**result} for data, result in zip(self._data_list, self._result_list)]
            return [single_output]
        else:
            success_output = [{**data,**result} for data, result, status in zip(self._data_list, self._result_list, self._status_list) if status["message"] is None]
            error_output = [{**data,**result} for data, result, status in zip(self._data_list, self._result_list, self._status_list) if status["message"] is not None]
            return [success_output, error_output]
    
    def write_result_to_output(self):
        while self.next_bundle_to_sent != self._bundle_number:
            time.sleep(0.01)
        self.output_function(self.get_final_data(single_list = self.single_output))
        self.next_bundle_to_sent = self.next_bundle_to_sent + 1
    
    def get_bundle_number(self):
        return self._bundle_number
    
    def get_status(self):
        return self._status_list

    def get_payload(self):
        return self._payload_list
      
    def get_rest_response(self):
        return self._rest_response

    def get_postprocess_result(self):
        return self._result_list


    #######################################################
    # to be defined by derived class
    #
    # they are specific for different asynch calls
    #######################################################        
    def preprocess(self):
        raise NotImplementedError

    def postprocess(self):
        raise NotImplementedError

    def synch_rest_call(self):
        raise NotImplementedError




# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2019

"""
Changes
+++++++

v1.0.3:  
- documentation updates: links in description updated

v1.0.1:     
- deprecate: parameter expected_load, but will be supported for backward 
             compatibility if new parameter bundle_size is not provided
- new: parameter bundles_size determines the bundle size directly


Overview
++++++++

Provides functions to use Watson Machine Learning (WML) online scoring in `topology based streaming applications <https://www.ibm.com/support/knowledgecenter/SSQNUZ_3.0.1/cpd/svc/streams/developing-intro.html>`_.
All models which are supported by WML for online scoring can be used to score streaming data in a topology application.
The models have to be created with tools provided by WML and Cloud Pak for Data. They need to be stored in WML repository 
and published as an online deployment. 


This package is designed for high tuple throughput. It is using the WML feature of sending multiple input data 
within a single scoring request to the online scoring REST endpoint. This minimizes the communication overhead. 
On the other side it will increase latency, but only in milli second region.


Quick ref to use this package
+++++++++++++++++++++++++++++

1. Create a model and store it in CP4D WML repository
   There are good tutorials available for the different supported ML frameworks.
   To be able to store a model in CP4D WML repository you need to create a 'deployment space'.
   You can create different spaces for different scenarioa (test, dev, production). Depending on 
   the tool you are using to create and store your model you will be asked to choose or create a 
   'deployment space'. In Jupyter notebook you need to assign the 'space id' to your created WML client.
   You can also assign a space to your project, which can now be used as 'deployment space' in the notebook.
   
       `CPD 3.0.1: Create and store models <https://www.ibm.com/support/knowledgecenter/SSQNUZ_3.0.1/wsj/analyze-data/ml-notebook_local.html>`_
   

2. Create a WML online scoring deployment
   You create a deployment by choosing a stored model ID and the type of deployment. Again you can do it 
   via CP4D GUI as context specific action on a shown model or via Python API e.g. in a notebook.
   You will get information about the progress and the final state of the deployment creation.
   An WML online scoring deployment is a REST endpoint to which you send (https request) your input data and get scoring results
   back as response.
   
       `CPD 3.0.1: Store and deploy models <https://www.ibm.com/support/knowledgecenter/SSQNUZ_3.0.1/wsj/wmls/wmls-deploy-python.html>`_
   

3. Create the streaming application
   Once the deployment is active you can create the streaming application using the streamsx Python package. 
   You can read data from several sources, transform them, do analytics and write them sinks. In this process you
   can integrate now the streamsx.wml.wml_online_scoring() function which provides the high performance approach of
   WML online scoring. Which means that multiple input data are send within a single REST request to the scoring REST
   endpoint. The maximum number of contained input data can be controlled by parameter.
   
       `CPD 3.0.1: Analyze streaming data <https://www.ibm.com/support/knowledgecenter/SSQNUZ_3.0.1/cpd/svc/streams/developing-intro.html>`_


Sample
++++++

A simple sample which scores on an online deployment running a model created for the well know IRIS data 
set to predict the Iris species from size of the petal and sepal of an exemplar::

    from icpd_core import icpd_util
    from streamsx.topology.topology import Topology
    from streamsx.topology import context
    import json
    import time
    import streamsx.wml as wml
    import streamsx.wml.utils as wml_utils
    
    streams_instance_name = 'streams'
    cfg=icpd_util.get_service_instance_details(name=streams_instance_name, instance_type="streams")
    
    #field mapping supports JSON string and Python dict
    field_mapping_dict =[{"model_field":"Sepal.Length",
                          "tuple_field":"sepal_length"},
                         {"model_field":"Sepal.Width",
                          "tuple_field":"sepal_width"},
                         {"model_field":"Petal.Length",
                          "tuple_field":"petal_length"},
                         {"model_field":"Petal.Width",
                          "tuple_field":"petal_width"}]
    
    #field mapping can be of type dict or a JSON string reflecting same content
    field_mapping = json.dumps(field_mapping_dict)
    
    # credentials support JSON string or Python dict
    logged_in_users_wml_credentials = json.dumps(wml_utils.get_wml_credentials())  #token,url,instance_id,version
    
    topo = Topology(name="WMLOnlineScoring")
    
    class TestSource:
        def __init__(self, ):
            pass
        def __call__(self):
            # let it stream forever
            counter = 0
            while True:
                counter += 1
                time.sleep(0.01)
                #yield everytime same values, doesn't matter for test
                yield {"petal_length":1.4,
                       "petal_width":0.2,
                       "sepal_length":5.1,
                       "sepal_width":3.5,
                       "number" : counter}
    
    records = topo.source(TestSource())
    
    # 2 result streams are generated: 1st with successful scorings, 2nd with failed scorings because of invalid input
    scorings,invalids = wml.wml_online_scoring(records, #input stream
                                         '72a15621-5e2e-44b5-a245-6a0fabc5de1e', #deployment_guid
                                         field_mapping, 
                                         logged_in_users_wml_credentials, 
                                         'e34d2846-cc27-4e8a-80af-3d0f7021d0cb', #space_guid
                                         bundle_size = 10,
                                         queue_size = 1000, 
                                         threads_per_node = 1)
    
    # publish results as JSON
    scorings.publish(topic="ScoredRecords",schema=json,name="PublishScores")
    score_view = scorings.view(name="ScoredRecords", description="Sample of scored records")
    
    # Disable SSL certificate verification on test
    cfg[context.ConfigParams.SSL_VERIFY] = False
    
    # build and submit
    submission_result = context.submit('DISTRIBUTED', 
                                       topo, 
                                       cfg)

"""


__version__='1.0.3'

__all__ = ['wml_online_scoring']
from streamsx.wml._wml import wml_online_scoring




# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2020

import os
import streamsx.spl.op
import streamsx.spl.types
from streamsx.topology.schema import CommonSchema, StreamSchema
from streamsx.spl.types import rstring
import json

PRIVATE_WML_TOOLKIT_NAMESPACE = "com.ibm.streams.wml"
PRIVATE_WML_TOOLKIT_LOCATION = "spl/toolkit.wml/com.ibm.streams.wml"

def _add_toolkit(topo):
    # specific private toolkit used in this wrapper
    # info.xml is an optional element -- so we don't provide for our private toolkit
    # toolkit.xml is generated during topology/application build -- so we don't care
    # So the only think is to provide the correct TK structure containing just the 
    # python file defining the primitive_operator, operator xml is also generated by topology/build
    # the toolkit is part of this streamsx.wml python package
    streamsx.spl.toolkit.add_toolkit(topo, os.path.dirname(os.path.realpath(__file__))+'/'+PRIVATE_WML_TOOLKIT_LOCATION)


def wml_online_scoring( stream, 
                        deployment_guid, 
                        #mapping_function,
                        field_mapping,
                        credentials, 
                        space_guid, 
                        expected_load = 0, 
                        queue_size = 2000,  
                        threads_per_node = 2,
                        single_output = False,
                        node_count = 1, 
                        connectionConfiguration=None, 
                        name = None,
                        bundle_size = 100):
    """
    Scoring tuples received from input :py:class:`stream` using the online scoring endpoint of WML referenced by the :py:class:`deployment_guid`.
    Mapping from input attributes to the models minning fields is done by the :py:class:`field_mapping` parameter.


    Args:
        stream (:py:class:`topology_ref:streamsx.topology.topology.Stream`):         
            input stream of data to be scored
        deployment_guid (str): 
            GUID of the WML deplyoment to be used to score the input data
        field_mapping (list|str): 
            list or JSON string representing the
            mappings between model fields and input tuple elements
            [{'model_field':str, 'is_mandatory':bool,'tuple_field':str},{...}]
        credentials (dict|str): 
            dict or JSON string of WML credentials to be used
            {'url':<cp4d_cluster_url>,'token':<token_of authenticated_user>,'instance_id': 'wml_local','version':'2.5.0'}
        space_guid (str): 
            GUID of the used deployment space
        expected_load (int, optional): 
            depricated, use new parameter bundle_size !
            The expected tuple throughput which is
            used to determine a maximum number of tuple to be sent in one request to 
            WML online scoring, maximum number = expected_load / threads_per_node
            this value has most impact on throughput performance, defaults to None
        queue_size (int, optional): 
            The internal buffer size after which back-pressure happens, defaults to 2000
        threads_per_node (int, optional): 
            optional field to set the number of threads used to process the received tuples
            may increase throughput performance by using free resources while other thread(s) waiting
            for the asynchronous result from online scoring, defaults to 2 
        single_output (bool, optional): 
            optional field to define if error data and success data should be send 
            to one output (stream) or to 2 different outputs (streams), defaults to False
        node_count(int, optional):
            optional field giving the number of REST nodes which share the load, defaults to 1
        name (str, optional): 
            Give the resulting Streams operator a name of your choice
        bundle_size (int, optional): 
            optional field for setting maximum number of tuple to be sent in one scoring request
            this value has most impact on throughput performance, defaults to 100
        
    Returns:
        result_stream, error_stream(:py:class:`topology_ref:streamsx.topology.topology.Stream`, :py:class:`topology_ref:streamsx.topology.topology.Stream`):
            Returns streams for further processing depending on ''single_output'' setting.
            
        ''single_output'' = False
            :result_stream: Tuples as received on input extended by the scoring result field predictions of type dict
            
            :error_stream: Tuples as received on input extended by the error indication `scoring_error` of type dict
            
        ''single_output'' = True:
            :result_stream: all Tuples as received on input, extended by the scoring result field ''prediction'' of type dict
                                or by field ''prediction_error'' of type dict
            
            
            
            
    """
    # add private toolkit
    topology = stream.topology
    _add_toolkit(topology)
    
    #check input parameter

    # field_mapping has to be either a list
    if isinstance(field_mapping,list):
        field_mapping_json = json.dumps(field_mapping)
    # may be also a JSON array
    elif isinstance(field_mapping,str):
        try:
            test_load=json.loads(field_mapping)
            field_mapping_json = field_mapping
        except:
            raise Exception("wml_online_scoring() parameter 'field_mapping' is not a valid json")    
    else:
        raise Exception("wml_online_scoring() parameter 'field_mapping' has to be either a 'list' or a valid JSON string of mappings")
    

    # credentials has to be either dict
    if isinstance(credentials,dict):
        credentials_json = json.dumps(credentials)
    # or a JSON string
    elif isinstance(credentials,str):
        try:
            test_load=json.loads(credentials)
            credentials_json = credentials
        except:
            raise Exception("wml_online_scoring() parameter 'credentials' is not a valid json")
            
    else:
        raise Exception("wml_online_scoring() parameter 'credentials' has to be either a 'dict' or a valid JSON string of credentials")

    # either expected_load(depricated) or bundle_size(defaults to 100) have to be set
    # For backward compatibility!!!
    #     if expected_load is set use this and not bundle_size
    _bundle_size = 0
    if expected_load is 0: 
        if isinstance(bundle_size, int):
            _bundle_size = bundle_size
        else:
            raise Exception("wml_online_scoring() parameter 'bundle_size' has to be an integer")
    elif not isinstance(expected_load,int):
        raise Exception("Depricated wml_online_scoring() parameter 'expected_load' has to be an integer, consider using the new parameter 'bundle_size'")
    else:
        print ("wml_online_scoring() parameter 'expected_load' is depricated but still supported, consider using the new parameter 'bundle_size'")


    # create instance of wrapper class
    _op = _WMLOnlineScoring(stream = stream, 
                            deployment_guid = deployment_guid, 
                            #mapping_function = mapping_function,
                            field_mapping = field_mapping_json,
                            credentials = credentials_json, 
                            space_guid = space_guid, 
                            expected_load = expected_load,
                            queue_size = queue_size,  
                            threads_per_node = threads_per_node,
                            single_output = single_output, 
                            node_count = node_count,
                            connectionConfiguration = connectionConfiguration, 
                            name = name,
                            bundle_size = _bundle_size)

    # calling SPL operators will result anytime in schema based output streams
    # these need to be mapped back to the Python object Stream we received

    if single_output:
        return _op.outputs[0]
    else:
        return _op.outputs[0],_op.outputs[1]




class _WMLOnlineScoring(streamsx.spl.op.Invoke):
    def __init__(self, stream, 
                       deployment_guid, 
                       #mapping_function,
                       field_mapping,
                       credentials, 
                       space_guid, 
                       expected_load, 
                       queue_size, 
                       threads_per_node,
                       single_output,
                       node_count, 
                       connectionConfiguration, 
                       name,
                       bundle_size):

        topology = stream.topology
        kind="com.ibm.streams.wml::WMLOnlineScoring"
        inputs=[stream]
        schemas=[object,object]  #not [StreamSchema('tuple<blob __spl_po>'),StreamSchema('tuple<blob __spl_po>')]
        params = dict()
        params['deployment_guid'] = deployment_guid
        params['wml_credentials'] = credentials
        #params['mapping_function'] = mapping_function
        params['field_mapping'] = field_mapping,
        params['space_guid'] = space_guid
        params['expected_load'] = expected_load
        params['queue_size'] = queue_size
        params['threads_per_node'] = threads_per_node
        params['single_output'] = single_output
        params['node_count'] = node_count
        params['bundle_size'] = bundle_size

        super(_WMLOnlineScoring, self).__init__(topology,kind,inputs,schemas,params,name)







#
#*******************************************************************************
#* Copyright (C) 2018, International Business Machines Corporation. 
#* All Rights Reserved. *
#*******************************************************************************
#
# WML specific imports
from watson_machine_learning_client import WatsonMachineLearningAPIClient
# REST handler specific imports
from .bundleresthandler import BundleRestHandler
# standard python imports
import logging
import json
import time
import threading
import pickle
import sys

#define tracer and logger
#logger for error which should and can! be handled by an administrator
#tracer for all other events that are of interest for developer
tracer = logging.getLogger(__name__)
    
logger = logging.getLogger("com.ibm.streams.log")


class BundleController():
    """
    
    """
    def __init__(self, 
                       expected_load = 0,    #depricated
                       queue_size = None, 
                       threads_per_node = None,
                       single_output = None,
                       node_count = None,
                       handler_class = None,
                       field_mapping = None,
                       output_function = None,
                       bundle_size = None
                      ):

        tracer.debug("__init__ called")

        #############################################################
        # Parameters
        ############################################################
        self._expected_load = expected_load
        self._max_queue_size = queue_size
        self._threads_per_node = threads_per_node
        self._single_output = single_output
        self._node_count = node_count
        self._max_request_size = bundle_size if expected_load is 0 else int(expected_load/self._threads_per_node/self._node_count)
        self._handler_class = handler_class

        
        ############################################################
        # internal variables
        ############################################################
        self._input_queue = list([])
        self._sending_threads = []
        self._lock = threading.Condition() # changed to condition
        self._output_lock = threading.Lock()
        self._thread_finish_counter = 0

        
        assert(self._handler_class is not None)
        ############################################################
        # Configure the handler class to be used
        # with the handler base classes class parameters 
        ############################################################
        self._handler_class.max_copy_size = self._max_request_size
        self._handler_class.input_list_lock = self._lock
        self._handler_class.source_data_list = self._input_queue
        self._handler_class.single_output = self._single_output
        self._handler_class.field_mapping = json.loads(field_mapping)
        self._handler_class.output_function = output_function

        tracer.debug("__init__ finished")
        return
        
    def process_data(self, input_data):
        """It is called for every single input data
        It will be just stored in the input queue. On max queue size processing
        blocks and backpressure on the up-stream/sending_thread happens.
        """
        with self._lock:
            self._lock.wait_for(lambda : len(self._input_queue) <= self._max_queue_size)
            self._input_queue.append(input_data)
            self._lock.notify()


    def prepare(self):
        self._create_sending_threads()
    
    def run(self):
        self._start_sending_threads()
        
    def stop(self):
        self._end_sending_threads()

    def finish(self):
        self._join_sending_threads()

    def _change_thread_number(self,delta):
        return

   
    def _create_sending_threads(self):
        for count in range(self._threads_per_node * self._node_count):
            tracer.debug("Create thread")
            handler_instance = self._handler_class(count)
            thread_control = {'index':count,'run':True, 'handler':handler_instance}
            thread_control['thread'] = threading.Thread(target = handler_instance.run)
            self._sending_threads.append(thread_control)
            tracer.debug("Thread data: %s",str(thread_control))

    
    def _start_sending_threads(self):
        for thread_control in self._sending_threads:
            tracer.debug("Start sending thread %s",str(thread_control))
            thread_control['thread'].start()
    
    def _end_sending_threads(self):
        for thread_control in self._sending_threads:
            #thread_control['run'] = False
            thread_control['handler'].stop()
            
    def _join_sending_threads(self):
        tracer.debug("_join_sending_threads called during processing of operator stop.")
        
        # trigger threads to signal that they are ready
        # each will decrement by 1 if all are ready it's again 0
        #self._thread_finish_counter = len(self._sending_threads)
        #tracer.debug("Wait for %d threads to finish processing of buffers", len(self._sending_threads))
        
        # wait that the trigger becomes 0 and all threads left their task func
        #while self._thread_finish_counter > 0 : time.sleep(1.0)
        #tracer.debug("All threads finished processing of buffers")

        for thread_control in self._sending_threads:
            thread_control['thread'].join()
            tracer.debug("Thread %d joined.", thread_control['index'])





# coding=utf-8
# Licensed Materials - Property of IBM
# Copyright IBM Corp. 2020

from watson_machine_learning_client import WatsonMachineLearningAPIClient


def get_wml_credentials(token=None,url=None,version="3.0.0"):
    """Returns a credential object to be used for WML connection
    
    Arguments:
    token -- token which should be included in the credentials
             this token has to be a non-expiring token as it is
             used at Streams runtime and no user credentials shall
             be needed resp. stored
             (default: users CPD token from Jupyter environment)
    url -- url of the CPD cluster where WML is running
    version -- version of the CPD cluster where WML is running
               type str, defaults to "3.0.0"
                 
    Call it only in the notebook where the topology is created, not in Python
    code executed at Streams runtime.
    """
    from icpd_core import icpd_util
        
    # take the token from notebook environment 
    if token is None:
        token = icpd_util.icpd_token
    if url is None:
        url = "https://internal-nginx-svc:12443"
    # credetials as to be used in CPD >=2.5
    credentials = {
                   "url": url, # access the main CP4D proxy/the CP4D cluster root from internal side
                   "token": token,
                   "instance_id": "wml_local",
                   "version" : version
                  }
    #test client creation to fail already in notebook if there is something wrong
    #need object copy here as the function changes "instance_id" in the object and so it 
    #couldn't be used anymore afterwards
    #wml_client = WatsonMachineLearningAPIClient(copy.copy(credentials))
    return credentials
    
def get_project_space(credentials):
    """Returns the notebooks project space GUID.
       
    Argument:
    credentials -- the credentials to be used to connect to WML
      
    Call it only in the notebook where the topology is created, not at Streams runtime.
    Models and deployments are placed in projects space if no other space is given at 
    their creation time.
    The space GUID is needed to instantiate a WMLOnlineScoring object.
    """
    from project_lib import Project
        
    wml_client = WatsonMachineLearningAPIClient(copy.copy(credentials))
    spaces = wml_client.spaces.get_details()["resources"]
    project = Project.access()
    project_guid = project.get_metadata()["metadata"]["guid"]
    # get the space associated with the project
    project_space=None
    for space in spaces:
        for tag in space["entity"]["tags"]:
            if tag["value"]=="dsx-project."+project_guid:
                project_space = space["metadata"]["id"]
    return project_space

